/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.plugin.jbiplugin;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuildingException;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Component;
import org.ow2.petals.jbi.descriptor.original.generated.Identification;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.Component.SharedLibrary;
import org.ow2.petals.plugin.jbiplugin.util.ZipUtil;
import org.ow2.petals.plugin.jbiplugin.util.ZipUtil.ZipEntryCallback;

/**
 * Configure a deliverable JBI Artefact (Component / Service Assembly)
 * 
 * @goal jbi-configure
 * @description Configure a deleverable JBI Artefact (Component / Service
 *              Assembly)
 * @since 2.0.0
 * 
 * @author cdeneux - Capgemini Sud
 */
public class JBIConfigureMojo extends JBIAbstractConfigurableMojo {

    /**
     * GroupId of the deliverable JBI artefact to configure.
     * 
     * @parameter expression="${groupId}"
     * @editable true
     * @required
     * @description GroupId of the deliverable JBI artefact to configure.
     */
    protected String groupId;

    /**
     * ArtifactId of the deliverable JBI artefact to configure.
     * 
     * @parameter expression="${artifactId}"
     * @editable true
     * @required
     * @description ArtifactId of the deliverable JBI artefact to configure.
     */
    protected String artifactId;

    /**
     * Version of the deliverable JBI artefact to configure.
     * 
     * @parameter expression="${version}"
     * @editable true
     * @required
     * @description Version of the deliverable JBI artefact to configure.
     */
    protected String version;

    /**
     * Configured artifacts output directory
     * 
     * @parameter expression="${project.build.directory}/configured-artifacts"
     * @editable true
     * @required
     * @description Configured artifacts output directory
     */
    protected File configuredArtifactOutputDirectory;

    /**
     * MavenProject of the provided artifact
     */
    private MavenProject providedProject;

    /**
     * the goal "jbi-configure" is designed to be run on command line. So we
     * need to override this method to avoid a check on packaging.
     * 
     * @see org.ow2.petals.plugin.jbiplugin.JBIAbstractMojo#packagingIsOkForGoal()
     */
    @Override
    protected boolean packagingIsOkForGoal() {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.plugin.jbiplugin.JBIAbstractMojo#executeMojo()
     */
    @Override
    public void executeMojo() throws MojoExecutionException, MojoFailureException {

        this.readConfigurationFiles();

        try {
            if (!this.configuredArtifactOutputDirectory.exists()) {
                this.configuredArtifactOutputDirectory.mkdirs();
            }

            // We are looking for POM file of the provided artifact
            Artifact artifact = this.artifactFactory.createProjectArtifact(this.groupId,
                    this.artifactId, this.version);
            this.providedProject = this.mavenProjectBuilder.buildFromRepository(artifact,
                    this.project.getRemoteArtifactRepositories(), this.localRepository);

            // We update with the correct packaging
            artifact = this.artifactFactory.createArtifact(this.groupId, this.artifactId,
                    this.version, null, this.providedProject.getPackaging());

            this.artifactResolver.resolve(artifact, this.project.getRemoteArtifactRepositories(),
                    this.localRepository);

            final String artifactType = this.providedProject.getPackaging();
            if (JBIAbstractMojo.PACKAGING_COMPONENT.equals(artifactType)) {
                this.configureComponent(artifact);
            } else if (JBIAbstractMojo.PACKAGING_SA.equals(artifactType)) {
                throw new MojoExecutionException("Coming next ...");
            } else if (JBIAbstractMojo.PACKAGING_SL.equals(artifactType)) {
                this.configureSharedLibrary(artifact);
            } else {
                throw new MojoExecutionException("Unsupported artifact type: " + artifactType);
            }
        } catch (final ArtifactResolutionException e) {
            throw new MojoExecutionException("A problem occurs during configuration.", e);
        } catch (final ArtifactNotFoundException e) {
            throw new MojoExecutionException("A problem occurs during configuration.", e);
        } catch (final ProjectBuildingException e) {
            throw new MojoExecutionException("A problem occurs during configuration.", e);
        }
    }

    private void configureComponent(final Artifact artifact) throws MojoExecutionException {

        try {
            final File artifactFile = artifact.getFile();
            this.info("The following file is used as artifact source: "
                    + artifactFile.getAbsolutePath());
            final ZipFile zipArtifactFile = new ZipFile(artifactFile);
            final File configuredArtifactFile = new File(this.configuredArtifactOutputDirectory,
                    artifact.getFile().getName());
            this.debug("Output JBI archive file: " + configuredArtifactFile.getAbsolutePath());
            final ZipOutputStream zipConfiguredArtifactFile = new ZipOutputStream(
                    new FileOutputStream(configuredArtifactFile));

            ZipUtil.copyAndUpdateZipFile(zipArtifactFile, zipConfiguredArtifactFile,
                    new ZipEntryCallback() {
                        public InputStream onZipEntry(final ZipEntry zipEntry,
                                final InputStream zipEntryInputStream) throws IOException,
                                MojoExecutionException {
                            if (JBIDESCRIPTOR_ZIPENTRY_NAME.equals(zipEntry.getName())) {
                                // The ZIP entry is the JBI descriptor, we
                                // update it
                                try {
                                    final Jbi jbiDescriptor = JBIDescriptorBuilder
                                            .buildJavaJBIDescriptor(zipEntryInputStream);

                                    final Component component = jbiDescriptor.getComponent();

                                    // We update the JBI component identifier
                                    // according to the
                                    // configuration
                                    final Identification ident = component.getIdentification();
                                    ident.setName(JBIConfigureMojo.this.evaluateJBIIdentifier(ident
                                            .getName(), artifact));

                                    // We update the shared-libraries list
                                    // according to the
                                    // configuration
                                    final List<SharedLibrary> sharedLibraries = JBIConfigureMojo.this
                                            .evaluateSharedLibraries(artifact);
                                    component.getSharedLibraryList().clear();
                                    for (final SharedLibrary sharedLibrary : sharedLibraries) {
                                        component.getSharedLibraryList().add(sharedLibrary);
                                    }

                                    final String jbiXml = JBIDescriptorBuilder
                                            .buildXmlStringJBIdescriptor(jbiDescriptor);
                                    // Reset of the JBI descriptor file time
                                    zipEntry.setTime(System.currentTimeMillis());
                                    return new ByteArrayInputStream(jbiXml.getBytes());
                                } catch (final JBIDescriptorException e) {
                                    throw new MojoExecutionException(
                                            "Error reading JBI descriptor.", e);
                                }
                            } else {
                                return zipEntryInputStream;
                            }
                        }
                    });

            zipConfiguredArtifactFile.flush();
            zipConfiguredArtifactFile.close();

            this.info("The configured JBI archive is available here: "
                    + configuredArtifactFile.getAbsolutePath());
        } catch (final IOException e) {
            throw new MojoExecutionException("A problem occurs during configuration.", e);
        }
    }

    private void configureSharedLibrary(final Artifact artifact) throws MojoExecutionException {
        try {
            final File artifactFile = artifact.getFile();
            this.info("The following file is used as artifact source: "
                    + artifactFile.getAbsolutePath());
            final ZipFile zipArtifactFile = new ZipFile(artifactFile);
            final File configuredArtifactFile = new File(this.configuredArtifactOutputDirectory,
                    artifact.getFile().getName());
            this.debug("Output JBI archive file: " + configuredArtifactFile.getAbsolutePath());
            final ZipOutputStream zipConfiguredArtifactFile = new ZipOutputStream(
                    new FileOutputStream(configuredArtifactFile));

            ZipUtil.copyAndUpdateZipFile(zipArtifactFile, zipConfiguredArtifactFile,
                    new ZipEntryCallback() {
                        public InputStream onZipEntry(final ZipEntry zipEntry,
                                final InputStream zipEntryInputStream) throws IOException,
                                MojoExecutionException {
                            if (JBIDESCRIPTOR_ZIPENTRY_NAME.equals(zipEntry.getName())) {
                                // The ZIP entry is the JBI descriptor, we
                                // update it
                                try {
                                    final Jbi jbiDescriptor = JBIDescriptorBuilder
                                            .buildJavaJBIDescriptor(zipEntryInputStream);

                                    final org.ow2.petals.jbi.descriptor.original.generated.Jbi.SharedLibrary sharedLibrary = jbiDescriptor
                                            .getSharedLibrary();

                                    // We update the JBI shared library
                                    // identifier according to the
                                    // configuration
                                    final Identification ident = sharedLibrary.getIdentification();
                                    ident.setName(JBIConfigureMojo.this.evaluateJBIIdentifier(ident
                                            .getName(), artifact));

                                    // We update the JBI shared library version
                                    // according to the
                                    // configuration
                                    sharedLibrary.setVersion(JBIConfigureMojo.this
                                            .evaluateJBIVersion(sharedLibrary.getVersion(),
                                                    artifact));

                                    final String jbiXml = JBIDescriptorBuilder
                                            .buildXmlStringJBIdescriptor(jbiDescriptor);
                                    // Reset of the JBI descriptor file time
                                    zipEntry.setTime(System.currentTimeMillis());
                                    return new ByteArrayInputStream(jbiXml.getBytes());
                                } catch (final JBIDescriptorException e) {
                                    throw new MojoExecutionException(
                                            "Error reading JBI descriptor.", e);
                                }
                            } else {
                                return zipEntryInputStream;
                            }
                        }
                    });

            zipConfiguredArtifactFile.flush();
            zipConfiguredArtifactFile.close();

        } catch (final IOException e) {
            throw new MojoExecutionException("A problem occurs during configuration.", e);
        }
    }

    @Override
    protected MavenProject getMavenProject() {
        return this.providedProject;
    }
}
