/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.plugin.jbiplugin;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.Set;
import java.util.zip.ZipException;
import java.util.zip.ZipOutputStream;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.artifact.versioning.InvalidVersionSpecificationException;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.Exclusion;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuildingException;
import org.codehaus.plexus.util.interpolation.MapBasedValueSource;
import org.codehaus.plexus.util.interpolation.ObjectBasedValueSource;
import org.codehaus.plexus.util.interpolation.RegexBasedInterpolator;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Component;
import org.ow2.petals.jbi.descriptor.original.generated.Consumes;
import org.ow2.petals.jbi.descriptor.original.generated.Identification;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.Provides;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.jbi.descriptor.original.generated.Target;
import org.ow2.petals.jbi.descriptor.original.generated.Component.SharedLibrary;
import org.ow2.petals.plugin.jbiplugin.util.PropertiesInterpolationValueSource;
import org.ow2.petals.plugin.jbiplugin.util.ZipUtil;
import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Make a JBI Archive of a project
 * 
 * @goal jbi-package
 * @description Make a JBI archive from a JBI component project
 * @requiresDependencyResolution runtime
 * 
 * @author alouis - EBM Websourcing
 * @author cdeneux - Capgemini Sud
 * @author rnaudin - EBM Websourcing
 */
public class JBIPackageMojo extends JBIAbstractConfigurableMojo {

    /**
     * The name of the Jar package of the JBI artifact
     * 
     * @parameter expression=
     *            "${project.build.directory}/${project.artifactId}-${project.version}.jar"
     * @editable true
     * @required
     * @description Name of the Jar package of the JBI artifact
     */
    protected File artifactJar;

    /**
     * Update JBI XML
     * 
     * @parameter expression="true"
     * @editable true
     * @required
     * @description Boolean which active the update of the JBI descriptor fields
     *              against the maven parameters
     */
    protected boolean updateJBIXml;

    /**
     * Sets the mapping pattern for all SU file names included in this service
     * assembly.
     * 
     * @parameter expression="$${artifactId}-$${version}.$${extension}"
     * @editable true
     * @required
     * @description Sets the mapping pattern for all SU file names included in
     *              this service assembly. Example of usage (caution to the
     *              usage of "$" and "$$"):
     *              <serviceUnitFileNameMappingInSA>$${artifactId
     *              }.$${extension}</serviceUnitFileNameMappingInSA>. Some extra
     *              pattern are defined: "version.major" is the major part of
     *              the version.
     */
    protected String serviceUnitFileNameMappingInSA;

    /**
     * If the JBI update is activated, sets the mapping pattern of the all SU
     * names included in this service assembly.
     * 
     * @parameter expression="$${artifactId}-$${version}"
     * @editable true
     * @required
     * @description If the JBI update is activated, sets the mapping pattern for
     *              all SU names included in this service assembly. Example of
     *              usage (caution to the usage of "$" and "$$"):
     *              <serviceUnitNameMappingInSA
     *              >$${artifactId}</serviceUnitNameMappingInSA>. Some extra
     *              pattern are defined: "version.major" is the major part of
     *              the version.
     */
    protected String serviceUnitNameMappingInSA;

    /**
     * If the JBI update is activated, sets the mapping pattern of the Service
     * Assembly name.
     * 
     * @parameter expression="$${artifactId}-$${version}"
     * @editable true
     * @required
     * @description If the JBI update is activated, sets the mapping pattern of
     *              the Service Assembly. Example of usage (caution to the usage
     *              of "$" and "$$"):
     *              <serviceAssemblyNameMapping>$${artifactId}<
     *              /serviceAssemblyNameMapping>. Some extra pattern are
     *              defined: "version.major" is the major part of the version.
     */
    protected String serviceAssemblyNameMapping;

    /**
     * If the JBI update is activated, sets the mapping pattern of the component
     * name.
     * 
     * @parameter expression="$${artifactId}"
     * @editable true
     * @required
     * @description If the JBI update is activated, sets the mapping pattern of
     *              the component name. Example of usage (caution to the usage
     *              of "$" and "$$"):
     *              <componentNameMapping>$${artifactId}</componentNameMapping>.
     *              Some extra pattern are defined: "version.major" is the major
     *              part of the version.
     */
    protected String componentNameMapping;

    /**
     * If the JBI update is activated, sets the mapping pattern of the Shared
     * Library name.
     * 
     * @parameter expression="$${artifactId}"
     * @editable true
     * @required
     * @description If the JBI update is activated, sets the mapping pattern of
     *              the Shared Library name. Example of usage (caution to the
     *              usage of "$" and "$$"):
     *              <sharedLibraryNameMapping>$${artifactId
     *              }</sharedLibraryNameMapping>. Some extra pattern are
     *              defined: "version.major" is the major part of the version.
     */
    protected String sharedLibraryNameMapping;

    /**
     * List of entries to exclude (comma separated values)
     * 
     * @parameter expression="foo"
     * @required
     * @editable true
     * @description List of entry to exclude (comma separated values)
     */
    protected String toExcludes;

    /**
     * Set of entries in the zip
     */
    private Set<String> entries = new HashSet<String>();

    /**
     * List of entries to exclude
     */
    private Set<String> entriesToExclude = new HashSet<String>();

    /**
     * Recurse a directory to add its content to a zip file. If updateJBIXml
     * file is true do not add the jbi.xml file to the archive (it will be
     * updated and added later).
     * 
     * @param zipOutputStream
     *            zip file
     * @param directory
     *            directory to recurse
     * @param entryDirectoryName
     *            directory to put the content in the zip
     * @throws Exception
     */
    private void recurseDirectory(ZipOutputStream zipOutputStream, File directory,
            String entryDirectoryName) throws IOException {
        for (File file : directory.listFiles()) {
            if (file.isFile()) {
                if (!updateJBIXml || !"jbi.xml".equals(file.getName())) {
                    this.info("     " + file.getName() + " added to JBI archive");
                    if (file.getName().equals("jbi.xml")) {
                        // Use "/" instead of File.separator to have a directory
                        // level in the ZIP file.
                        ZipUtil.addFile(zipOutputStream, new FileInputStream(file),
                                (entryDirectoryName.equals("") ? "" : entryDirectoryName + "/")
                                        + "META-INF/" + file.getName());
                    } else {
                        // Use "/" instead of File.separator to have a directory
                        // level in the ZIP file.
                        ZipUtil.addFile(zipOutputStream, new FileInputStream(file),
                                (entryDirectoryName.equals("") ? "" : entryDirectoryName + "/")
                                        + file.getName());
                    }
                }
            } else {
                this.debug("File entry: " + file.getName());
                // Avoid to add the .svn and CVS directories and add directory
                // to zip.
                if (!file.getName().startsWith(".")
                        && !file.getName().toUpperCase(Locale.getDefault()).equals("CVS")) {
                    // Use "/" instead of File.separator to have a directory
                    // level in the ZIP file
                    recurseDirectory(zipOutputStream, file, (entryDirectoryName.equals("") ? ""
                            : entryDirectoryName + "/")
                            + file.getName());
                }
            }
        }
    }

    /**
     * Extract the list of jar to exclude from the archive
     * 
     */
    private void extractExcludeValues() {
        // We remove the space
        if (!"foo".equalsIgnoreCase(toExcludes)) {
            toExcludes = toExcludes.replace(" ", "");
            for (String jar : toExcludes.split(",")) {
                entriesToExclude.add(jar);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.plugin.jbiplugin.JBIAbstractMojo#executeMojo()
     */
    @Override
    public void executeMojo() throws MojoExecutionException, MojoFailureException {

        this.debug("Configuration mode: "
                + (this.includeConfiguration ? "activated" : "unactivated"));
        if (this.includeConfiguration) {
            this.readConfigurationFiles();
        }

        this.attachedJBIArchive(project.getArtifact());

        extractExcludeValues();

        if (outputDirectory.exists()) {
            if (project.getArtifact().getFile().exists()) {
                project.getArtifact().getFile().delete();
            }
        } else {
            outputDirectory.mkdirs();
        }

        if (jbiDirectory.exists()) {
            if (PACKAGING_COMPONENT.equals(project.getPackaging())) {
                packageComponent();
            } else if (PACKAGING_SU.equals(project.getPackaging())) {
                packageServiceUnit();
            } else if (PACKAGING_SA.equals(project.getPackaging())) {
                packageServiceAssembly();
            } else if (PACKAGING_SL.equals(project.getPackaging())) {
                packageSharedLibrary();
            }
        } else {
            this.error("JBI directory [" + jbiDirectory + "] does not exist.");
            throw new MojoFailureException("JBI directory [" + jbiDirectory + "] does not exist.");
        }
    }

    /**
     * Create a service-unit package.
     */
    private void packageComponent() throws MojoExecutionException {
        try {
            this.info("Start building JBI component archive "
                    + project.getArtifact().getFile().getAbsolutePath());

            // Add the jbi files
            ZipOutputStream zipOutputStream = new ZipOutputStream(new FileOutputStream(project
                    .getArtifact().getFile()));
            recurseDirectory(zipOutputStream, jbiDirectory, "");
            zipOutputStream.flush();

            // Add the component jar
            addArtifactResources(zipOutputStream);

            // Add the component dependencies
            addDependencies(zipOutputStream);

            if (updateJBIXml) {
                // Update the JBI XML file
                updateComponentJBIXmlFile(zipOutputStream);
            }

            zipOutputStream.flush();
            zipOutputStream.close();

            this.info("JBI component archive building done.");
        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Create a service-unit package.
     */
    private void packageServiceUnit() throws MojoExecutionException {
        try {
            this.info("Start building JBI service-unit archive "
                    + project.getArtifact().getFile().getAbsolutePath());

            // JBI Archive creation
            ZipOutputStream zipOutputStream = new ZipOutputStream(new FileOutputStream(project
                    .getArtifact().getFile()));

            try {
                // Load the JBI descriptor
                InputStream jbiDescriptorToUpdateInputStream = new FileInputStream(jbiDirectory
                        + File.separator + "jbi.xml");
                Jbi jbiDescriptor = JBIDescriptorBuilder
                        .buildJavaJBIDescriptor(jbiDescriptorToUpdateInputStream);

                if (jbiDescriptor == null) {
                    throw new MojoExecutionException("JBI Descriptor is null");
                }

                InputStream jbiDescriptorInputStream;
                if (this.updateJBIXml) {
                    // Update SU extensions according to project properties
                    for (Consumes consumes : jbiDescriptor.getServices().getConsumes()) {
                        List<Element> nodeList = consumes.getAnyOrAny();
                        evaluateElementList(nodeList);
                    }
                    for (Provides provides : jbiDescriptor.getServices().getProvides()) {
                        List<Element> nodeList = provides.getAnyOrAny();
                        evaluateElementList(nodeList);
                    }

                    String jbiXml = JBIDescriptorBuilder.buildXmlStringJBIdescriptor(jbiDescriptor);
                    this.debug("Generated JBI Descriptor:\n" + jbiXml);
                    jbiDescriptorInputStream = new ByteArrayInputStream(jbiXml.getBytes());
                    // Add the jbi.xml file
                    ZipUtil.addFile(zipOutputStream, jbiDescriptorInputStream,
                            JBIDESCRIPTOR_ZIPENTRY_NAME);
                    this.info("\tJBI descriptor added to JBI Service Unit archive");
                }
            } catch (IOException e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            } catch (JBIDescriptorException e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            } catch (Throwable e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            }

            // Add the jbi files
            recurseDirectory(zipOutputStream, jbiDirectory, "");
            zipOutputStream.flush();

            // Add the service unit resources
            addArtifactResources(zipOutputStream);

            // Add the service unit dependencies
            addDependencies(zipOutputStream);

            zipOutputStream.close();

            this.info("JBI Service unit archive building done.");

        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Create a service-assembly package.
     */
    private void packageServiceAssembly() throws MojoExecutionException {

        this.info("Start building JBI service-assembly archive "
                + project.getArtifact().getFile().getAbsolutePath());

        try {

            // JBI Archive creation
            ZipOutputStream zipOutputStream = new ZipOutputStream(new FileOutputStream(project
                    .getArtifact().getFile()));

            try {
                // Load the JBI descriptor
                InputStream jbiDescriptorToUpdateInputStream = new FileInputStream(jbiDirectory
                        + File.separator + "jbi.xml");
                Jbi jbiDescriptor = JBIDescriptorBuilder
                        .buildJavaJBIDescriptor(jbiDescriptorToUpdateInputStream);

                if (jbiDescriptor == null) {
                    throw new MojoExecutionException("JBI Descriptor is null");
                }

                if (this.updateJBIXml) {
                    if (jbiDescriptor.getServiceAssembly() == null) {
                        throw new MojoExecutionException(
                                "No service-assembly in the JBI descriptor.");
                    }
                    if (jbiDescriptor.getServiceAssembly().getIdentification() == null) {
                        throw new MojoExecutionException("No identification in the JBI descriptor.");
                    }

                    String artifactName = this.evaluateJBIIdentifier(
                            this.serviceAssemblyNameMapping, this.project.getArtifact());

                    jbiDescriptor.getServiceAssembly().getIdentification().setName(artifactName);

                    jbiDescriptor.getServiceAssembly().getIdentification().setDescription(
                            this.project.getDescription());

                    jbiDescriptor.getServiceAssembly().getServiceUnit().clear();
                }

                // Add the service unit dependencies files
                List<Dependency> dependencies = project.getDependencies();
                for (Dependency dependency : dependencies) {

                    if (PACKAGING_SU.equals(dependency.getType())) {
                        Artifact serviceUnitArtifact = createDependencyArtifact(dependency);

                        this.artifactResolver.resolve(serviceUnitArtifact, this.project
                                .getRemoteArtifactRepositories(), this.localRepository);

                        MavenProject serviceUnitProject = this.mavenProjectBuilder
                                .buildFromRepository(serviceUnitArtifact, this.project
                                        .getRemoteArtifactRepositories(), this.localRepository);

                        final String zipDependencyFileName = serviceUnitArtifact.getFile()
                                .getName();

                        if (this.updateJBIXml) {

                            ServiceUnit serviceUnit = new ServiceUnit();
                            Identification identification = new Identification();
                            identification.setName(this.evaluateJBIIdentifier(
                                    this.serviceUnitNameMappingInSA, serviceUnitProject
                                            .getArtifact()));

                            identification.setDescription(serviceUnitProject.getDescription());
                            serviceUnit.setIdentification(identification);

                            // We find the target component name following
                            // dependencies
                            Jbi jbiComponentDescriptor = null;
                            Artifact serviceUnitComponentArtifact = null;
                            boolean foundInServiceUnitDependencies = false;
                            List<Dependency> serviceUnitDependencies = serviceUnitProject
                                    .getDependencies();
                            for (Dependency serviceUnitDependency : serviceUnitDependencies) {
                                this.debug("SU dependency: " + serviceUnitDependency);
                                if (PACKAGING_COMPONENT.equals(serviceUnitDependency.getType())) {
                                    // We have found the component
                                    this.debug("Component dependency found: "
                                            + serviceUnitDependency);
                                    foundInServiceUnitDependencies = true;

                                    try {

                                        serviceUnitComponentArtifact = createDependencyArtifact(serviceUnitDependency);
                                        this.artifactResolver.resolve(serviceUnitComponentArtifact,
                                                this.project.getRemoteArtifactRepositories(),
                                                this.localRepository);

                                        jbiComponentDescriptor = this
                                                .readJbiDescriptor(serviceUnitComponentArtifact);

                                        this.debug("Component found: "
                                                + serviceUnitComponentArtifact);

                                    } catch (InvalidVersionSpecificationException e) {
                                        throw new MojoExecutionException(e.getLocalizedMessage(), e);
                                    } catch (ArtifactResolutionException e) {
                                        throw new MojoExecutionException(e.getLocalizedMessage(), e);
                                    } catch (ArtifactNotFoundException e) {
                                        throw new MojoExecutionException(e.getLocalizedMessage(), e);
                                    }

                                    break;
                                }
                            }

                            if (!foundInServiceUnitDependencies) {
                                throw new MojoExecutionException(
                                        "No component defined in SU as dependency.");
                            }

                            this.debug("Component name: "
                                    + jbiComponentDescriptor.getComponent().getIdentification()
                                            .getName());
                            Target target = new Target();
                            // target.setComponentName(jbiComponentDescriptor.getComponent()
                            // .getIdentification().getName());

                            target.setComponentName(this.evaluateJBIIdentifier(
                                    jbiComponentDescriptor.getComponent().getIdentification()
                                            .getName(), serviceUnitComponentArtifact));

                            target.setArtifactsZip(zipDependencyFileName);
                            serviceUnit.setTarget(target);

                            jbiDescriptor.getServiceAssembly().getServiceUnit().add(serviceUnit);
                        }

                        this.debug("SU artifact file: " + serviceUnitArtifact.getFile());
                        ZipUtil.addFile(zipOutputStream, new FileInputStream(serviceUnitArtifact
                                .getFile()), zipDependencyFileName);

                        this.info("\t" + serviceUnitArtifact.getFile().getName()
                                + " added to JBI Service Assembly archive (as "
                                + zipDependencyFileName + ")");
                    }
                }

                InputStream jbiDescriptorInputStream;
                if (this.updateJBIXml) {
                    String jbiXml = JBIDescriptorBuilder.buildXmlStringJBIdescriptor(jbiDescriptor);
                    this.debug("Generated JBI Descriptor:\n" + jbiXml);
                    jbiDescriptorInputStream = new ByteArrayInputStream(jbiXml.getBytes());
                } else {
                    jbiDescriptorInputStream = new FileInputStream(jbiDirectory + File.separator
                            + "jbi.xml");
                }

                // Add the jbi.xml file
                ZipUtil.addFile(zipOutputStream, jbiDescriptorInputStream,
                        JBIDESCRIPTOR_ZIPENTRY_NAME);
                this.info("\tJBI descriptor added to JBI Service Assembly archive");

            } catch (IOException e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            } catch (JBIDescriptorException e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            } catch (ProjectBuildingException e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            } catch (InvalidVersionSpecificationException e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            } catch (Throwable e) {
                throw new MojoExecutionException(e.getLocalizedMessage(), e);
            } finally {
                try {
                    zipOutputStream.close();
                } catch (ZipException e) {
                    // This exception occurs on ZipOutputStream.close() when no
                    // entry has been added, so we ignore it
                } catch (IOException e) {
                    throw new MojoExecutionException(e.getLocalizedMessage(), e);
                }
            }

            this.info("JBI Service Assembly archive building done.");

        } catch (FileNotFoundException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Create a shared-library package.
     */
    private void packageSharedLibrary() throws MojoExecutionException {
        try {
            this.info("Start building JBI shared-library archive "
                    + project.getArtifact().getFile().getAbsolutePath());

            // JBI Archive creation
            ZipOutputStream zipOutputStream = new ZipOutputStream(new FileOutputStream(project
                    .getArtifact().getFile()));

            // Add the jbi files
            recurseDirectory(zipOutputStream, jbiDirectory, "");
            zipOutputStream.flush();

            // Add the shared-library jar
            addArtifactResources(zipOutputStream);

            // Add the shared library dependencies
            addDependencies(zipOutputStream);

            if (updateJBIXml) {
                // Update the JBI XML file
                updateSharedLibraryJBIXmlFile(zipOutputStream);
            }

            zipOutputStream.close();

            this.info("JBI Shared library archive building done.");

        } catch (IOException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Add the Java resources (classes and resources) in the JBI archive.
     * 
     * @param zipOutputStream
     *            The JBI archive
     */
    private void addArtifactResources(ZipOutputStream zipOutputStream) throws IOException {
        if (artifactJar.exists()) {
            entries.add(artifactJar.getName());
            zipFile(zipOutputStream, artifactJar);
        }
    }

    /**
     * Add required dependencies in the JBI archive. excluded dependencies are : <br>
     * test/provided scope, component/SharedLib type, and optional dependencies
     * 
     * @param zipOutputStream
     *            The JBI archive
     * @throws MojoExecutionException
     */
    private void addDependencies(ZipOutputStream zipOutputStream) throws IOException,
            MojoExecutionException {

        // Create the list of all artifact files for this project.
        // Some artifacts are not required:
        // - scope: test, provided
        // - package: jbi-component, jbi-shared-library
        // So, find dependencies to really include by analyzing dependency
        // graph.

        Set<Artifact> wholeArtifactsSet = project.getArtifacts();
        List<Dependency> dependencies = project.getDependencies();

        for (Dependency dependency : dependencies) {
            // include the dependency if required
            if (includeDependencyInPackage(dependency, false)) {
                debug("Required     : " + dependency.getArtifactId() + ":"
                        + dependency.getVersion());
                addDependency(dependency, wholeArtifactsSet, zipOutputStream,
                        new ArrayList<Exclusion>());
            } else {
                debug("Not required : " + dependency.getArtifactId() + ":"
                        + dependency.getVersion());
            }
        }
    }

    /**
     * Find transitive dependencies of the given dependency to be included
     * (recursive)
     * 
     * @param dependency
     *            dependency from which transitive dependencies to include are
     *            retrieved MUST be non null
     * @param wholeArtifactFiles
     *            whole artifacts list (even not required)
     * @param zipOutputStream
     *            Zip to write artifacts into
     * @param parentDependencyExclusions
     *            Dependencies exclusion list defined by parent dependencies in
     *            the tags &lt;exclude&gt;
     * @throws IOException
     *             problem while adding files in the Zip
     * @throws MojoExecutionException
     *             problem while ananlysing the dependencies
     */
    private void addDependency(final Dependency dependency, final Set<Artifact> wholeArtifactFiles,
            final ZipOutputStream zipOutputStream, final List<Exclusion> parentDependencyExclusions)
            throws IOException, MojoExecutionException {

        try {

            // create an artifact from the dependency
            final Artifact dependencyArtifact = createDependencyArtifact(dependency);
            this.artifactResolver.resolve(dependencyArtifact, this.project
                    .getRemoteArtifactRepositories(), this.localRepository);

            // Add the artifact to the Package (if found in the
            // wholeArtifactsList)
            if (wholeArtifactFiles.contains(dependencyArtifact)) {
                this.entries.add(dependencyArtifact.getFile().getName());
                this.zipFile(zipOutputStream, dependencyArtifact.getFile());
            } else {
                this.debug("Rejected     : " + dependencyArtifact.getArtifactId() + ":"
                        + dependencyArtifact.getVersion() + " (not selected by Maven)");
            }

            // find all the dependencies of this artifact
            // (create a mavenproject from the artifact )
            final MavenProject dependencyProject = buildProjectFromArtifact(dependencyArtifact);
            final List<Dependency> subDependencies = dependencyProject.getDependencies();
            List<Exclusion> dependencyExclusions = new ArrayList(parentDependencyExclusions);
            dependencyExclusions.addAll(dependency.getExclusions());

            for (final Dependency dependencyDependency : subDependencies) {
                // Check if the dependency is declared as excluded in POM
                boolean excluded = false;
                for (final Exclusion exclusion : dependencyExclusions) {
                    if (dependencyDependency.getGroupId().equals(exclusion.getGroupId())
                            && dependencyDependency.getArtifactId().equals(
                                    exclusion.getArtifactId())) {
                        excluded = true;
                        break;
                    }
                }

                if (excluded) {
                    this.debug("Not required : " + dependencyDependency.getArtifactId() + ":"
                            + dependencyDependency.getVersion() + " (dependency exclusion)");
                } else {

                    // Check if the dependency should be included
                    if (includeDependencyInPackage(dependencyDependency, true)) {
                        this.debug("Required     : " + dependencyDependency.getArtifactId() + ":"
                                + dependencyDependency.getVersion() + " (transitive dependency of "
                                + dependency.getArtifactId() + ":" + dependency.getVersion() + ")");
                        this.addDependency(dependencyDependency, wholeArtifactFiles,
                                zipOutputStream, dependencyExclusions);
                    } else {
                        this.debug("Not required : " + dependencyDependency.getArtifactId() + ":"
                                + dependencyDependency.getVersion() + " (transitive dependency of "
                                + dependency.getArtifactId() + ":" + dependency.getVersion() + ")");
                    }
                }
            }

        } catch (final InvalidVersionSpecificationException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        } catch (final ProjectBuildingException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        } catch (final ArtifactResolutionException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        } catch (final ArtifactNotFoundException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Clone a set of String
     * 
     * @param setToClone
     * @return
     */
    private Set<String> cloneSet(Set<String> setToClone) {
        Set<String> outSet = new HashSet<String>();
        for (String string : setToClone) {
            outSet.add(string);
        }
        return outSet;
    }

    /**
     * 
     * @param zipOutputStream
     * @throws MojoExecutionException
     */
    private void updateComponentJBIXmlFile(ZipOutputStream zipOutputStream)
            throws MojoExecutionException {
        try {
            final InputStream jbiDescriptorToUpdateInputStream = new FileInputStream(jbiDirectory
                    + File.separator + "jbi.xml");
            final Jbi jbiDescriptor = JBIDescriptorBuilder
                    .buildJavaJBIDescriptor(jbiDescriptorToUpdateInputStream);
            Component component = jbiDescriptor.getComponent();

            // set the component identification
            String artifactName = this.evaluateJBIIdentifier(this.componentNameMapping,
                    this.project.getArtifact());
            component.getIdentification().setName(artifactName);
            component.getIdentification().setDescription(this.project.getDescription());

            // set the component classpath according to dependencies
            Set<String> elemsToAdd = cloneSet(entries);
            final List<String> compoClassPath = component.getComponentClassPath().getPathElement();
            for (String elem : compoClassPath) {
                if (elemsToAdd.contains(elem)) {
                    elemsToAdd.remove(elem);
                }
            }
            compoClassPath.clear();
            compoClassPath.addAll(elemsToAdd);

            // set the bootstrap classpath according to dependencies
            elemsToAdd = cloneSet(entries);
            final List<String> bootClassPath = component.getBootstrapClassPath().getPathElement();
            for (String elem : bootClassPath) {
                if (elemsToAdd.contains(elem)) {
                    elemsToAdd.remove(elem);
                }
            }
            bootClassPath.clear();
            bootClassPath.addAll(elemsToAdd);

            // We update the shared-libraries list according to the
            // configuration and mapping rules
            final List<SharedLibrary> sharedLibraries = this.evaluateSharedLibraries(this.project
                    .getArtifact());
            component.getSharedLibraryList().clear();
            for (final SharedLibrary sharedLibrary : sharedLibraries) {
                component.getSharedLibraryList().add(sharedLibrary);
            }

            final String jbiXml = JBIDescriptorBuilder.buildXmlStringJBIdescriptor(jbiDescriptor);
            final InputStream jbiDescriptorInputStream = new ByteArrayInputStream(jbiXml.getBytes());
            ZipUtil.addFile(zipOutputStream, jbiDescriptorInputStream, JBIDESCRIPTOR_ZIPENTRY_NAME);

        } catch (Exception e) {
            throw new MojoExecutionException(
                    "Problem while updating jbi.xml file. We add it like it is", e);
        }
    }

    /**
     * Update the JBI xml file to added all the dependencies of a shared library
     * JBI archive.
     * 
     * @param zipOutputStream
     * @throws MojoExecutionException
     */
    private void updateSharedLibraryJBIXmlFile(ZipOutputStream zipOutputStream)
            throws MojoExecutionException {
        try {
            final InputStream jbiDescriptorToUpdateInputStream = new FileInputStream(jbiDirectory
                    + File.separator + "jbi.xml");
            final Jbi jbiDescriptor = JBIDescriptorBuilder
                    .buildJavaJBIDescriptor(jbiDescriptorToUpdateInputStream);

            org.ow2.petals.jbi.descriptor.original.generated.Jbi.SharedLibrary sharedLibrary = jbiDescriptor
                    .getSharedLibrary();

            // add SL name if not already specified
            final String artifactName = this.evaluateJBIIdentifier(this.sharedLibraryNameMapping,
                    this.project.getArtifact());
            sharedLibrary.getIdentification().setName(artifactName);
            // add SL version if not already specified
            sharedLibrary.setVersion(this.evaluateJBIVersion(this.project.getVersion(),
                    this.project.getArtifact()));

            // Set the SL classpath
            final Set<String> elemsToAdd = cloneSet(entries);
            final List<String> slClassPath = sharedLibrary.getSharedLibraryClassPath()
                    .getPathElement();
            for (String elem : slClassPath) {
                if (elemsToAdd.contains(elem)) {
                    elemsToAdd.remove(elem);
                }
            }
            elemsToAdd.add(artifactJar.getName());
            slClassPath.clear();
            slClassPath.addAll(elemsToAdd);

            final String jbiXml = JBIDescriptorBuilder.buildXmlStringJBIdescriptor(jbiDescriptor);
            final InputStream jbiDescriptorInputStream = new ByteArrayInputStream(jbiXml.getBytes());
            ZipUtil.addFile(zipOutputStream, jbiDescriptorInputStream, JBIDESCRIPTOR_ZIPENTRY_NAME);

        } catch (Exception e) {
            throw new MojoExecutionException(
                    "Problem while updating jbi.xml file. We add it like it is", e);
        }
    }

    /**
     * Add a file to a zip stream, if not already zipped
     * 
     * @param zipOutputStream
     * @param file
     */
    private void zipFile(ZipOutputStream zipOutputStream, File file) {

        try {
            ZipUtil.addFile(zipOutputStream, new FileInputStream(file), file.getName());
            this.info("\t" + file.getName() + " added to JBI archive");
        } catch (final IOException e) {
            entriesToExclude.add(file.getName());
            this.error("Problem while adding " + file.getName() + " to JBI archive", e);
        }
    }

    /**
     * Evaluates Filename Mapping (source code copied from
     * maven-assembly-plugin)
     * 
     * @param expression
     * @param artifact
     * @return mapped expression
     * @throws AssemblyFormattingException
     * @throws org.apache.maven.plugin.MojoExecutionException
     */
    // TODO (CDE): Rework to simplify after fixing of MNG-3558
    private String evaluateFileNameMapping(final String expression, final Artifact artifact,
            final Properties properties) {

        this.debug("Evaluating expression: " + expression);

        String value = expression.replaceAll("\\$\\{dollarSign\\}", "\\$");

        // FIXME: This is BAD! Accessors SHOULD NOT change the behavior of the
        // object.
        artifact.isSnapshot();

        RegexBasedInterpolator interpolator = new RegexBasedInterpolator();

        if (artifact != null) {
            interpolator.addValueSource(new ObjectBasedValueSource(artifact));
            interpolator.addValueSource(new ObjectBasedValueSource(artifact.getArtifactHandler()));
        }
        if (properties != null) {
            interpolator.addValueSource(new MapBasedValueSource(properties));
        }

        Properties classifierMask = new Properties();
        classifierMask.setProperty("classifier", "");

        interpolator.addValueSource(new PropertiesInterpolationValueSource(classifierMask));

        value = interpolator.interpolate(value, "__artifact");

        this.debug("Expression evaluated to: " + value);

        return value;
    }

    /**
     * Evaluate the given nodeList.
     * 
     * @param nodeList
     * @throws DOMException
     */
    private void evaluateElementList(final List<Element> nodeList) throws DOMException {

        for (Element elt : nodeList) {
            evaluateNodeList(elt.getChildNodes());
        }
    }

    private void evaluateNodeList(final NodeList nodeList) throws DOMException {

        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() == Node.TEXT_NODE
                    && !nodeList.item(i).getTextContent().trim().equals("")) {
                String mappedText = evaluateFileNameMapping(nodeList.item(i).getTextContent(),
                        this.project.getArtifact(), this.project.getProperties());
                nodeList.item(i).setTextContent(mappedText);
            } else if (nodeList.item(i).getChildNodes() != null) {
                evaluateNodeList(nodeList.item(i).getChildNodes());
            }
        }
    }

    /**
     * <p>
     * Return true if the dependency has to be included in the JBI package.
     * Excluded dependencies are :
     * <ul>
     * <li>scope test</li>
     * <li>scope provided</li>
     * <li>type PACKAGING_SL</li>
     * <li>type PACKAGING_COMPONENT</li>
     * <li>transitive and optional</li>
     * <li>already included</li>
     * </ul>
     * </p>
     * 
     * @param dependency
     *            MUST be non null
     * @param isTransitiveDependency
     *            true is the <code>dependency</code> is a transitive
     *            dependency.
     * @return true if the dependency has to be included in the package
     */
    private boolean includeDependencyInPackage(final Dependency dependency,
            final boolean isTransitiveDependency) throws MojoExecutionException {

        try {

            final String scope = dependency.getScope();
            if (scope != null
                    && (Artifact.SCOPE_TEST.equals(scope) || Artifact.SCOPE_PROVIDED.equals(scope))) {
                this.debug(dependency.getGroupId() + "-" + dependency.getArtifactId()
                        + "'s scope not required");
                return false;
            }

            if (isTransitiveDependency && dependency.isOptional()) {
                this.debug(dependency.getGroupId() + "-" + dependency.getArtifactId()
                        + " is optional and transitive, so not required");
                return false;
            }

            // create an artifact from the dependency
            final Artifact dependencyArtifact = this.createDependencyArtifact(dependency);
            this.artifactResolver.resolve(dependencyArtifact, this.project
                    .getRemoteArtifactRepositories(), this.localRepository);

            final String artifactName = dependencyArtifact.getFile().getName();

            if (this.entries.contains(artifactName)) {
                this.debug(artifactName + " already marked as required");
                return false;
            }

            if (this.entriesToExclude.contains(artifactName)) {
                this.debug(artifactName + " is marked as excluded");
                return false;
            }

            final String type = dependency.getType();
            if (type != null
                    && (JBIAbstractMojo.PACKAGING_SL.equals(type) || JBIAbstractMojo.PACKAGING_COMPONENT
                            .equals(type))) {
                this.debug(artifactName + "'s type not required");
                return false;
            }

            return true;
        } catch (final ArtifactResolutionException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        } catch (final ArtifactNotFoundException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        } catch (final InvalidVersionSpecificationException e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    @Override
    protected MavenProject getMavenProject() {
        return this.project;
    }
}
